c-----------------------------------------------------------
c     Data structure for FAS multi-grid program
c-----------------------------------------------------------

c-----------------------------------------------------------
c     Main storage array; all arrays used in MG solve are 
c     "allocated" from 'q'.  The integer 'qptr' maintains 
c     the index into 'q' representing the "high-water mark"
c     of allocation. See routine 'gsinit()'.
c
c     Make sure that this array is "sensibly" dimensioned 
c     for the machine you are running on.  On SGI machines 
c     you can type 'hinv' to find out how much physical
c     memory is available; the limit for regular Fortran
c     storage will less than or equal to physical memory
c     and sometimes substantially less.  On einstein
c     the limit *is* the amount of physical memory, 
c     but you should avoid running with that much.
c
c     Also note that 2**20 = 1M, so that 2**20 real*8 
c     variables is 8M of storage.
c-----------------------------------------------------------
      integer         qsize 
      parameter   ( qsize = 2**25 + 2**16)

      real*8        q(qsize)
      common      / comr8 /      q

      integer       qptr
      common      / comi4 /      qptr

      integer       lmax
      parameter   ( lmax = 12 )    

c-----------------------------------------------------------
c     Pointers for various grid functions.   Once these 
c     arrays have been initialized (see routine 'gsinit'),
c     q(u(l)), for example, is the start of storage for 
c     the level-l 'unknown' grid function.  This 
c     implementation uses temporary arrays ('t1', 't2', 't3') 
c     heavily in order to  keep the code relatively 
c     straightforward.  A production code (particularly a 
c     3-d code) should be more economical of storage. 
c-----------------------------------------------------------
      integer       u(lmax),     rhs(lmax),     tau(lmax),   
     &              t1(lmax),    t2(lmax),      t3(lmax)

      common      / comi4 /      
     &              u,           rhs,           tau,
     &              t1,          t2,            t3

c-----------------------------------------------------------
c     Added an excision mask function.
c     Since we will NOT be using full-weighted/half-weighted
c     restriction, we will need to write new operators for
c     the mask.  Thus we might as well make it an integer.
c
c     Note that we simply "hard code" the storage for the 
c     mask, rather than using something like the "qalloc" 
c     routine.
c-----------------------------------------------------------
c Storage for the mask
      integer       qmsize
      parameter   ( qmsize = 2**24+1 )

      integer       qm(qmsize)
      common      / comi4 /      qm

c Pointer to the mask data
      integer       mask(lmax)
      common      / comi4 /      mask

c Flags for excised, (normal) interior, and "extra smoothing" (interior) points
      integer       m_ex
      parameter   ( m_ex   = 1 )
      integer       m_int
      parameter   ( m_int = 0 )
      integer       m_esr
      parameter   ( m_esr = 2 )
      integer       m_esr_pass
      parameter   ( m_esr_pass = -2 )


c-----------------------------------------------------------
c     Mesh spacings and grid sizes.
c-----------------------------------------------------------
      real*8        h(lmax) 
      common     /  comr8 /      h

      integer       nx(lmax),    ny(lmax)
      common      / comi4 /      
     &              nx,          ny

c-----------------------------------------------------------
c     Coarse grid convergence criterion and iteration limit
c-----------------------------------------------------------
      real*8        epsi1
      parameter   ( epsi1  = 1.0d-10 )

      integer       maxsweep1
      parameter   ( maxsweep1 = 100 )
