c======================================================================
c  Utilities for managing values inside the excision mask
c  Scott Hawley
c  October 31, 2002
c======================================================================

c======================================================================
c     minit: Initialize the mask function
c
c     This will initialize the mask function with two shapes,
c     e.g. two circles, which are on either side of the y axis
c======================================================================
      subroutine minit(maxlev, lmin, mrad, mdisp_x)
            
      implicit none
     
      include 'mg.inc'
               
      integer     maxlev, lmin
      integer     mpcount
      integer     lastval 
            
c  Physical Radius of mask on finest grid.
c  Note that negative value turns excision off.
      real*8     mrad

c  x-displacement of center of each mask shape".  Choosing
c  mdisp_x = 0 will give only one excision region
      real*8     mdisp_x


      integer     l
         
      lastval = 1
      mask(lmin) = 1
c      write(*,*) '      level,  2**l,     2*h,  '
c     &        //'   npoints,    2*"area", N(l)/N(l-1)'
      do l = lmin, maxlev
        write(0,*)' Setting mask on level ',l
        if (l .gt. lmin) mask(l) = mask(l-1) + (2**(l-1) + 1)**2
         call imls  ( qm(mask(l)), m_int, nx(l), ny(l) )
         call mset  ( qm(mask(l)), nx(l), ny(l), h(l), mrad, 
     &                mdisp_x, mpcount, m_ex, m_int, m_esr, 
     &                m_esr_pass, l-1 )
c         write(*,1000) l, 2**l, 1.0d0/2**l, mpcount,
c     &               4*mpcount*(1.0d0/2**l)**2, mpcount*1.0d0/lastval
1000     format (2i10, f12.7, i10,2f12.7)
        if (l .le. 4) then
             call mprint( qm(mask(l)), nx(l), ny(l))
        endif
        lastval = mpcount
      enddo

      return
      end



c======================================================================
c     mset: This actually sets the values in the mask
c======================================================================
      subroutine  mset(mask, nx, ny, h, mrad, mdisp_x, mpcount,
     &                 m_ex, m_int, m_esr, m_esr_pass, esr_size)
      implicit none

      integer  nx,            ny
      integer  mask(nx, ny)
      real*8  h
      real*8   mrad,          mdisp_x
      integer  m_ex,          m_int
c    Used for "extra smoothing" region
      integer  m_esr,         m_esr_pass,        esr_size
      integer  mpcount

      integer  i, j
      integer  loop
      real*8   rad
      real*8   x,             y
      real*8   morig_x,           morig_y

      logical     msquare
      parameter ( msquare = .false. )

      mpcount = 0


c    Two "holes"
      morig_x = mdisp_x
      morig_y = 0.0d0
      do loop = 1, 2
         if (.not. msquare) then
            do i = 1, nx
               do j = 1, ny
                  rad =  sqrt (
     &             (h*(i - (nx/2+1)) - morig_x)**2 +
     &             (h*(j - (ny/2+1)) - morig_y)**2 )
                  if ( rad .le. mrad) then
                     mask(i,j) = m_ex
                     mpcount = mpcount + 1
                  endif
               enddo
            enddo
         else
            do i = 1, nx
               do j = 1, ny
                   x = (i - (nx/2+1) )*h - morig_x
                   y = (j - (ny/2+1) )*h - morig_y
                   if ( abs(x) .le. mrad  .and. abs(y) .le. mrad) then
                     mask(i,j) = m_ex
                     mpcount = mpcount + 1
                   endif
               enddo
            enddo
         endif
         morig_x = - morig_x
      enddo

c---------------------------------------------------------------
c    Set "extra smoothing" region around excision regions
c---------------------------------------------------------------
      do loop = 1, esr_size
        do i = 2, nx-1
           do j = 2, ny-1
              if ( (mask(i,j) .eq. m_int)  .and. (
     &              mask(i+1,j) .eq. m_ex .or.
     &              mask(i-1,j) .eq. m_ex .or.
     &              mask(i,j+1) .eq. m_ex .or.
     &              mask(i,j-1) .eq. m_ex .or.
     &              mask(i+1,j) .eq. m_esr .or.
     &              mask(i-1,j) .eq. m_esr .or.
     &              mask(i,j+1) .eq. m_esr .or.
     &              mask(i,j-1) .eq. m_esr) ) then

c            We need to set it it m_esr_pass instead of m_esr
c            because otherwise the loop with set EVERY following
c            point to m_esr.

                  mask(i,j) = m_esr_pass
              endif
           enddo
        enddo

c            THEN we go back and set the m_esr_pass's to m_esr
        do i = 2, nx-1
           do j = 2, ny-1
              if (mask(i,j) .eq. m_esr_pass) mask(i,j) = m_esr
           enddo
        enddo
      enddo

      return
      end

c===========================================================
c     Supplies exact solution on mask:
c
c        u(x,y) = sin(l_x pi x) sin(l_y pi y)
c===========================================================
      subroutine exonm(u,mask,h,nx,ny,m_ex)

         implicit       none

         include       'mgprob.inc'

         integer        nx,       ny
         real*8         u(nx,ny)
         integer        mask(nx,ny)
         integer        m_ex
         real*8         h

         integer        i,        j

         do  j = 1 , ny
            do  i = 1 , nx
               if (mask(i,j) .eq. m_ex)
     &            u(i,j) = sin(pi * lx * (i - 1) * h) *
     &                     sin(pi * ly * (j - 1) * h)
            end do
         end do
c         call dmlbs(u,0.0d0,nx,ny)

         return

      end

c======================================================================
c     mprint: handy routine for showing the mask
c======================================================================
      subroutine mprint(mask, nx, ny)
      implicit none

      integer nx, ny
      integer mask(nx, ny)

      integer i,j

      write(0,*) 'Mask:'
      do j = 1, ny
        write(0,1000) ( mask(i,j), i = 1, nx )
      enddo
1000  format(200I2)
      write(0,*)

      return
      end



c===========================================================
c     Load 2-D integer array with scalar.
c===========================================================
      subroutine imls(a,s,d1,d2)

         implicit       none

         integer        d1,       d2
         integer        a(d1,d2)
         integer        s

         integer        i,        j

         do j = 1 , d2
            do i = 1 , d1
               a(i,j) = s
            end do     
         end do

         return

      end

